import datetime
import json

import pytest

from pulp_glue.common.openapi import OpenAPI, _Response

pytestmark = pytest.mark.glue

TEST_SCHEMA = json.dumps(
    {
        "openapi": "3.0.3",
        "paths": {"test/": {"get": {"operationId": "test_id"}}},
        "components": {"schemas": {}},
    }
).encode()


@pytest.fixture
def openapi(monkeypatch: pytest.MonkeyPatch) -> OpenAPI:
    monkeypatch.setattr(OpenAPI, "load_api", lambda self, refresh_cache: TEST_SCHEMA)
    openapi = OpenAPI("base_url", "doc_path")
    openapi._parse_api(TEST_SCHEMA)
    return openapi


def test_parse_response_returns_dict_for_no_content(openapi: OpenAPI) -> None:
    response = _Response(204, {}, b"")
    result = openapi._parse_response({}, response)
    assert result == {}


def test_parse_response_decodes_json(openapi: OpenAPI) -> None:
    response = _Response(200, {"content-type": "application/json"}, b'{"a": 1, "b": "Hallo!"}')
    result = openapi._parse_response(
        {"responses": {"200": {"content": {"application/json": {}}}}}, response
    )
    assert result == {"a": 1, "b": "Hallo!"}


def test_extract_params_with_no_match(openapi: OpenAPI) -> None:
    parameters = {"a": 1, "b": "C"}
    query_params = openapi._extract_params("query", {}, {}, parameters)
    assert query_params == {}
    assert parameters == {"a": 1, "b": "C"}


def test_extract_params_raises_for_missing_required_parameter(openapi: OpenAPI) -> None:
    parameters = {"a": 1, "b": "C"}
    with pytest.raises(RuntimeError, match=r"Required parameters \[c\] missing"):
        openapi._extract_params(
            "query",
            {"parameters": [{"name": "c", "in": "query", "required": True}]},
            {},
            parameters,
        )
    assert parameters == {"a": 1, "b": "C"}


def test_extract_params_removes_matches(openapi: OpenAPI) -> None:
    parameters = {"a": 1, "c": "C"}
    query_params = openapi._extract_params(
        "query", {"parameters": [{"name": "c", "in": "query", "required": True}]}, {}, parameters
    )
    assert query_params == {"c": "C"}
    assert parameters == {"a": 1}


def test_extract_params_encodes_date(openapi: OpenAPI) -> None:
    parameters = {"a": datetime.date(2000, 1, 1)}
    query_params = openapi._extract_params(
        "query",
        {
            "parameters": [
                {"name": "a", "in": "query", "schema": {"type": "string", "format": "date"}}
            ]
        },
        {},
        parameters,
    )
    assert query_params == {"a": "2000-01-01"}



# copyright (c) 2020, Matthias Dellweg
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import json
import os
import typing as t
from collections import defaultdict
from dataclasses import dataclass
from io import BufferedReader
from urllib.parse import urlencode, urljoin

import requests
import urllib3
from multidict import CIMultiDict, MutableMultiMapping

from pulp_glue.common import __version__
from pulp_glue.common.exceptions import (
    OpenAPIError,
    PulpAuthenticationFailed,
    PulpException,
    PulpHTTPError,
    PulpNotAutorized,
    UnsafeCallError,
    ValidationError,
)
from pulp_glue.common.i18n import get_translation
from pulp_glue.common.schema import encode_json, encode_param, validate

translation = get_translation(__package__)
_ = translation.gettext

UploadType = t.Union[bytes, t.IO[bytes]]

SAFE_METHODS = ["GET", "HEAD", "OPTIONS"]


@dataclass
class _Response:
    status_code: int
    headers: t.Union[MutableMultiMapping[str], t.MutableMapping[str, str]]
    body: bytes


class AuthProviderBase:
    """
    Base class for auth providers.

    This abstract base class will analyze the authentication proposals of the openapi specs.
    Different authentication schemes should be implemented by subclasses.
    Returned auth objects need to be compatible with `requests.auth.AuthBase`.
    """

    def basic_auth(self, scopes: t.List[str]) -> t.Optional[requests.auth.AuthBase]:
        """Implement this to provide means of http basic auth."""
        return None

    def http_auth(
        self, security_scheme: t.Dict[str, t.Any], scopes: t.List[str]
    ) -> t.Optional[requests.auth.AuthBase]:
        """Select a suitable http auth scheme or return None."""
        # https://www.iana.org/assignments/http-authschemes/http-authschemes.xhtml
        if security_scheme["scheme"] == "basic":
            result = self.basic_auth(scopes)
            if result:
                return result
        return None

    def oauth2_client_credentials_auth(
        self, flow: t.Any, scopes: t.List[str]
    ) -> t.Optional[requests.auth.AuthBase]:
        """Implement this to provide other authentication methods."""
        return None

    def oauth2_auth(
        self, security_scheme: t.Dict[str, t.Any], scopes: t.List[str]
    ) -> t.Optional[requests.auth.AuthBase]:
        """Select a suitable oauth2 flow or return None."""
        # Check flows by preference.
        if "clientCredentials" in security_scheme["flows"]:
            flow = security_scheme["flows"]["clientCredentials"]
            # Select this flow only if it claims to provide all the necessary scopes.
            # This will allow subsequent auth proposals to be considered.
            if set(scopes) - set(flow["scopes"]):
                return None

            result = self.oauth2_client_credentials_auth(flow, scopes)
            if result:
                return result
        return None

    def __call__(
        self,
        security: t.List[t.Dict[str, t.List[str]]],
        security_schemes: t.Dict[str, t.Dict[str, t.Any]],
    ) -> t.Optional[requests.auth.AuthBase]:

        # Reorder the proposals by their type to prioritize properly.
        # Select only single mechanism proposals on the way.
        proposed_schemes: t.Dict[str, t.Dict[str, t.List[str]]] = defaultdict(dict)
        for proposal in security:
            if len(proposal) == 0:
                # Empty proposal: No authentication needed. Shortcut return.
                return None
            if len(proposal) == 1:
                name, scopes = list(proposal.items())[0]
                proposed_schemes[security_schemes[name]["type"]][name] = scopes
            # Ignore all proposals with more than one required auth mechanism.

        # Check for auth schemes by preference.
        if "oauth2" in proposed_schemes:
            for name, scopes in proposed_schemes["oauth2"].items():
                result = self.oauth2_auth(security_schemes[name], scopes)
                if result:
                    return result

        # if we get here, either no-oauth2, OR we couldn't find creds
        if "http" in proposed_schemes:
            for name, scopes in proposed_schemes["http"].items():
                result = self.http_auth(security_schemes[name], scopes)
                if result:
                    return result

        raise OpenAPIError(_("No suitable auth scheme found."))


class BasicAuthProvider(AuthProviderBase):
    """
    Implementation for AuthProviderBase providing basic auth with fixed `username`, `password`.
    """

    def __init__(self, username: str, password: str):
        self.auth = requests.auth.HTTPBasicAuth(username, password)

    def basic_auth(self, scopes: t.List[str]) -> t.Optional[requests.auth.AuthBase]:
        return self.auth


class OpenAPI:
    """
    The abstraction Layer to interact with a server providing an openapi v3 specification.

    Parameters:
        base_url: The base URL inlcuding the HTTP scheme, hostname and optional subpaths of the
            served api.
        doc_path: Path of the json api doc schema relative to the `base_url`.
        headers: Dictionary of additional request headers.
        auth_provider: Object that returns requests auth objects according to the api spec.
        cert: Client certificate used for auth.
        key: Matching key for `cert` if not already included.
        verify: Whether to check server TLS certificates agains a CA (requests semantic).
        refresh_cache: Whether to fetch the api doc regardless.
        safe_calls_only: Flag to disallow issuing POST, PUT, PATCH or DELETE calls.
        debug_callback: Callback that will be called with strings useful for logging or debugging.
        user_agent: String to use in the User-Agent header.
        cid: Correlation ID to send with all requests.
    """

    def __init__(
        self,
        base_url: str,
        doc_path: str,
        headers: t.Optional[CIMultiDict[str]] = None,
        auth_provider: t.Optional[AuthProviderBase] = None,
        cert: t.Optional[str] = None,
        key: t.Optional[str] = None,
        verify: t.Optional[t.Union[bool, str]] = True,
        refresh_cache: bool = False,
        safe_calls_only: bool = False,
        debug_callback: t.Optional[t.Callable[[int, str], t.Any]] = None,
        user_agent: t.Optional[str] = None,
        cid: t.Optional[str] = None,
    ):
        self._debug_callback: t.Callable[[int, str], t.Any] = debug_callback or (lambda i, x: None)
        self._base_url: str = base_url
        self._doc_path: str = doc_path
        self._safe_calls_only: bool = safe_calls_only
        self._headers = CIMultiDict(headers or {})
        self._verify = verify
        self._auth_provider = auth_provider
        self._cert = cert
        self._key = key

        self._headers.update(
            {
                "User-Agent": user_agent or f"Pulp-glue openapi parser ({__version__})",
                "Accept": "application/json",
            }
        )
        if cid:
            self._headers["Correlation-Id"] = cid

        self._setup_session()

        self.load_api(refresh_cache=refresh_cache)

    def _setup_session(self) -> None:
        # This is specific requests library.

        if self._verify is False:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        self._session: requests.Session = requests.session()
        # Don't redirect, because carrying auth accross redirects is unsafe.
        self._session.max_redirects = 0
        self._session.headers.update(self._headers)
        if self._auth_provider:
            if self._cert or self._key:
                raise OpenAPIError(_("Cannot use both 'auth' and 'cert'."))
        else:
            if self._cert and self._key:
                self._session.cert = (self._cert, self._key)
            elif self._cert:
                self._session.cert = self._cert
            elif self._key:
                raise OpenAPIError(_("Cert is required if key is set."))
        session_settings = self._session.merge_environment_settings(
            self._base_url, {}, None, self._verify, None
        )
        self._session.verify = session_settings["verify"]
        self._session.proxies = session_settings["proxies"]

    @property
    def base_url(self) -> str:
        return self._base_url

    @property
    def cid(self) -> t.Optional[str]:
        return self._headers.get("Correlation-Id")

    def load_api(self, refresh_cache: bool = False) -> None:
        # TODO: Find a way to invalidate caches on upstream change
        xdg_cache_home: str = os.environ.get("XDG_CACHE_HOME") or "~/.cache"
        apidoc_cache: str = os.path.join(
            os.path.expanduser(xdg_cache_home),
            "squeezer",
            (self._base_url + "_" + self._doc_path).replace(":", "_").replace("/", "_")
            + "api.json",
        )
        try:
            if refresh_cache:
                raise IOError()
            with open(apidoc_cache, "rb") as f:
                data: bytes = f.read()
            self._parse_api(data)
        except Exception:
            # Try again with a freshly downloaded version
            data = self._download_api()
            self._parse_api(data)
            # Write to cache as it seems to be valid
            os.makedirs(os.path.dirname(apidoc_cache), exist_ok=True)
            with open(apidoc_cache, "bw") as f:
                f.write(data)

    def _parse_api(self, data: bytes) -> None:
        self.api_spec: t.Dict[str, t.Any] = json.loads(data)
        if self.api_spec.get("openapi", "").startswith("3."):
            self.openapi_version: int = 3
        else:
            raise OpenAPIError(_("Unknown schema version"))
        self.operations: t.Dict[str, t.Any] = {
            method_entry["operationId"]: (method, path)
            for path, path_entry in self.api_spec["paths"].items()
            for method, method_entry in path_entry.items()
            if method in {"get", "put", "post", "delete", "options", "head", "patch", "trace"}
        }

    def _download_api(self) -> bytes:
        try:
            response: requests.Response = self._session.get(urljoin(self._base_url, self._doc_path))
        except requests.RequestException as e:
            raise OpenAPIError(str(e))
        response.raise_for_status()
        if "Correlation-Id" in response.headers:
            self._set_correlation_id(response.headers["Correlation-Id"])
        return response.content

    def _set_correlation_id(self, correlation_id: str) -> None:
        if "Correlation-Id" in self._headers:
            if self._headers["Correlation-Id"] != correlation_id:
                raise OpenAPIError(
                    _("Correlation ID returned from server did not match. {} != {}").format(
                        self._headers["Correlation-Id"], correlation_id
                    )
                )
        else:
            self._headers["Correlation-Id"] = correlation_id
            # Do it for requests too...
            self._session.headers["Correlation-Id"] = correlation_id

    def param_spec(
        self, operation_id: str, param_type: str, required: bool = False
    ) -> t.Dict[str, t.Any]:
        method, path = self.operations[operation_id]
        path_spec = self.api_spec["paths"][path]
        method_spec = path_spec[method]

        param_spec = {
            entry["name"]: entry
            for entry in path_spec.get("parameters", [])
            if entry["in"] == param_type
        }
        param_spec.update(
            {
                entry["name"]: entry
                for entry in method_spec.get("parameters", [])
                if entry["in"] == param_type
            }
        )
        if required:
            param_spec = {k: v for k, v in param_spec.items() if v.get("required", False)}
        return param_spec

    def _extract_params(
        self,
        param_in: str,
        path_spec: t.Dict[str, t.Any],
        method_spec: t.Dict[str, t.Any],
        params: t.Dict[str, t.Any],
    ) -> t.Dict[str, t.Any]:
        param_specs = {
            entry["name"]: entry
            for entry in path_spec.get("parameters", [])
            if entry["in"] == param_in
        }
        param_specs.update(
            {
                entry["name"]: entry
                for entry in method_spec.get("parameters", [])
                if entry["in"] == param_in
            }
        )
        result: t.Dict[str, t.Any] = {}
        for name in list(params.keys()):
            if name in param_specs:
                param = params.pop(name)
                param_spec = param_specs.pop(name)
                param_schema = param_spec.get("schema")
                if param_schema is not None:
                    self.validate_schema(param_schema, name, param)

                param = encode_param(param)
                if isinstance(param, list):
                    if not param:
                        # Don't propagate an empty list here
                        continue
                    # Check if we need to implode the list
                    style = (
                        param_spec.get("style") or "form"
                        if param_in in ("query", "cookie")
                        else "simple"
                    )
                    explode: bool = param_spec.get("explode", style == "form")
                    if not explode:
                        # Not exploding means comma separated list
                        param = ",".join(param)
                result[name] = param
        remaining_required = [
            item["name"] for item in param_specs.values() if item.get("required", False)
        ]
        if any(remaining_required):
            raise RuntimeError(
                _("Required parameters [{required}] missing in {param_in}.").format(
                    required=", ".join(remaining_required), param_in=param_in
                )
            )
        return result

    def validate_schema(self, schema: t.Any, name: str, value: t.Any) -> None:
        validate(schema, name, value, self.api_spec["components"]["schemas"])

    def _render_request_body(
        self,
        method_spec: t.Dict[str, t.Any],
        body: t.Optional[t.Dict[str, t.Any]] = None,
        validate_body: bool = True,
    ) -> t.Tuple[
        t.Optional[str],
        t.Optional[t.Union[t.Dict[str, t.Any], str]],
        t.Optional[t.List[t.Tuple[str, t.Tuple[str, UploadType, str]]]],
    ]:
        content_types: t.List[str] = []
        try:
            request_body_spec = method_spec["requestBody"]
        except KeyError:
            if body is not None:
                raise OpenAPIError(_("This operation does not expect a request body."))
            return None, None, None
        else:
            body_required = request_body_spec.get("required", False)
            if body is None and not body_required:
                # shortcut
                return None, None, None
            content_types = list(request_body_spec["content"].keys())
        assert body is not None

        content_type: t.Optional[str] = None
        data: t.Optional[t.Union[t.Dict[str, t.Any], str]] = None
        files: t.Optional[t.List[t.Tuple[str, t.Tuple[str, UploadType, str]]]] = None

        candidate_content_types = [
            "multipart/form-data",
        ]
        if not any((isinstance(value, (bytes, BufferedReader)) for value in body.values())):
            candidate_content_types = [
                "application/json",
                "application/x-www-form-urlencoded",
            ] + candidate_content_types
        errors: t.List[str] = []
        for candidate in candidate_content_types:
            content_type = next(
                (
                    content_type
                    for content_type in content_types
                    if content_type.startswith(candidate)
                ),
                None,
            )
            if content_type:
                if validate_body:
                    try:
                        self.validate_schema(
                            request_body_spec["content"][content_type]["schema"],
                            "body",
                            body,
                        )
                    except ValidationError as e:
                        errors.append(f"{content_type}: {e}")
                        # Try the next content-type
                        continue

                if content_type.startswith("application/json"):
                    data = encode_json(body)
                elif content_type.startswith("application/x-www-form-urlencoded"):
                    data = body
                elif content_type.startswith("multipart/form-data"):
                    uploads: t.Dict[str, t.Tuple[str, UploadType, str]] = {}
                    data = {}
                    # Extract and prepare the files to upload
                    if body:
                        for key, value in body.items():
                            if isinstance(value, (bytes, BufferedReader)):
                                uploads[key] = (
                                    getattr(value, "name", key),
                                    value,
                                    "application/octet-stream",
                                )
                            else:
                                data[key] = value
                    if uploads:
                        files = [(key, upload_data) for key, upload_data in uploads.items()]
                break
        else:
            # No known content-type left
            if errors:
                raise ValidationError(
                    _("Validation failed for '{operation_id}':\n  ").format(
                        operation_id=method_spec["operationId"]
                    )
                    + "\n  ".join(errors)
                )
            else:
                raise ValidationError(_("No valid content type found."))

        return content_type, data, files

    def _send_request(
        self,
        path_spec: t.Dict[str, t.Any],
        method: str,
        url: str,
        params: t.Dict[str, t.Any],
        headers: t.Dict[str, str],
        body: t.Optional[t.Dict[str, t.Any]] = None,
        validate_body: bool = True,
    ) -> _Response:
        method_spec = path_spec[method]
        content_type, data, files = self._render_request_body(method_spec, body, validate_body)
        security: t.Optional[t.List[t.Dict[str, t.List[str]]]] = method_spec.get(
            "security", self.api_spec.get("security")
        )
        if security and self._auth_provider:
            if "Authorization" in self._session.headers:
                # Bad idea, but you wanted it that way.
                auth = None
            else:
                auth = self._auth_provider(security, self.api_spec["components"]["securitySchemes"])
        else:
            # No auth required? Don't provide it.
            # No auth_provider available? Hope for the best (should do the trick for cert auth).
            auth = None
        # For we encode the json on our side.
        # Somehow this does not work properly for multipart...
        if content_type is not None and content_type.startswith("application/json"):
            headers["content-type"] = content_type
        request = self._session.prepare_request(
            requests.Request(
                method,
                url,
                auth=auth,
                params=params,
                headers=headers,
                data=data,
                files=files,
            )
        )
        if content_type:
            assert request.headers["content-type"].startswith(
                content_type
            ), f"{request.headers['content-type']} != {content_type}"
        for key, value in request.headers.items():
            self._debug_callback(2, f"  {key}: {value}")
        if request.body is not None:
            self._debug_callback(3, f"{request.body!r}")
        if self._safe_calls_only and method.upper() not in SAFE_METHODS:
            raise UnsafeCallError(_("Call aborted due to safe mode"))
        try:
            response = self._session.send(request)
        except requests.TooManyRedirects as e:
            assert e.response is not None
            raise OpenAPIError(
                _("Received redirect to '{url}'. Please check your CLI configuration.").format(
                    url=e.response.headers["location"]
                )
            )
        except requests.RequestException as e:
            raise OpenAPIError(str(e))
        self._debug_callback(
            1, _("Response: {status_code}").format(status_code=response.status_code)
        )
        for key, value in response.headers.items():
            self._debug_callback(2, f"  {key}: {value}")
        if response.text:
            self._debug_callback(3, f"{response.text}")
        if "Correlation-Id" in response.headers:
            self._set_correlation_id(response.headers["Correlation-Id"])
        if response.status_code == 401:
            raise PulpAuthenticationFailed(method_spec["operationId"])
        if response.status_code == 403:
            raise PulpNotAutorized(method_spec["operationId"])
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if e.response is not None:
                raise PulpHTTPError(str(e.response.text), e.response.status_code)
            else:
                raise PulpException(str(e))
        return _Response(
            status_code=response.status_code, headers=response.headers, body=response.content
        )

    def _parse_response(self, method_spec: t.Dict[str, t.Any], response: _Response) -> t.Any:
        if response.status_code == 204:
            return {}

        try:
            response_spec = method_spec["responses"][str(response.status_code)]
        except KeyError:
            # Fallback 201 -> 200
            try:
                response_spec = method_spec["responses"][str(100 * int(response.status_code / 100))]
            except KeyError:
                raise OpenAPIError(
                    _("Unexpected response '{code}' (expected '{expected}').").format(
                        code=response.status_code,
                        expected=(", ").join(method_spec["responses"].keys()),
                    )
                )
        content_type = response.headers.get("content-type")
        if content_type is not None and content_type.startswith("application/json"):
            assert content_type in response_spec["content"]
            return json.loads(response.body)
        return None

    def call(
        self,
        operation_id: str,
        parameters: t.Optional[t.Dict[str, t.Any]] = None,
        body: t.Optional[t.Dict[str, t.Any]] = None,
        validate_body: bool = True,
    ) -> t.Any:
        """
        Make a call to the server.

        Parameters:
            operation_id: ID of the operation in the openapi v3 specification.
            parameters: Arguments that are to be sent as headers, querystrings or part of the URI.
            body: Body payload for POST, PUT, PATCH calls.
            validate_body: Indicate whether the body should be validated.

        Returns:
            The JSON decoded server response if any.

        Raises:
            OpenAPIValidationError: on failed input validation (no request was sent to the server).
            requests.HTTPError: on failures related to the HTTP call made.
        """
        method, path = self.operations[operation_id]
        path_spec = self.api_spec["paths"][path]
        method_spec = path_spec[method]

        if parameters is None:
            parameters = {}
        else:
            parameters = parameters.copy()

        if any(self._extract_params("cookie", path_spec, method_spec, parameters)):
            raise NotImplementedError(_("Cookie parameters are not implemented."))

        headers = self._extract_params("header", path_spec, method_spec, parameters)

        for name, value in self._extract_params("path", path_spec, method_spec, parameters).items():
            path = path.replace("{" + name + "}", value)

        query_params = self._extract_params("query", path_spec, method_spec, parameters)

        if any(parameters):
            raise OpenAPIError(
                _("Parameter [{names}] not available for {operation_id}.").format(
                    names=", ".join(parameters.keys()), operation_id=operation_id
                )
            )
        url = urljoin(self._base_url, path)

        if query_params:
            qs = urlencode(query_params)
            self._debug_callback(1, f"{operation_id} : {method} {url}?{qs}")
        else:
            self._debug_callback(1, f"{operation_id} : {method} {url}")
        self._debug_callback(
            2, "\n".join([f"  {key}=={value}" for key, value in query_params.items()])
        )

        response = self._send_request(
            path_spec,
            method,
            url,
            query_params,
            headers,
            body,
            validate_body=validate_body,
        )

        return self._parse_response(method_spec, response)



import random
import string
import typing as t

import pytest

from pulp_glue.common.context import PulpContext
from pulp_glue.file.context import PulpFileRepositoryContext

pytestmark = pytest.mark.glue


@pytest.fixture
def file_repository(pulp_ctx: PulpContext) -> t.Iterator[t.Dict[str, t.Any]]:
    name = "".join(random.choices(string.ascii_letters, k=8))
    file_repository_ctx = PulpFileRepositoryContext(pulp_ctx)
    yield file_repository_ctx.create(body={"name": name})
    file_repository_ctx.delete()


@pytest.mark.live
def test_entity_list(pulp_ctx: PulpContext, file_repository: t.Dict[str, t.Any]) -> None:
    entity_ctx = PulpFileRepositoryContext(pulp_ctx)
    assert (
        len(entity_ctx.list(limit=1, offset=0, parameters={"name": file_repository["name"]})) == 1
    )


@pytest.mark.live
def test_entity_list_iterator(pulp_ctx: PulpContext, file_repository: t.Dict[str, t.Any]) -> None:
    entity_ctx = PulpFileRepositoryContext(pulp_ctx)
    stats: t.Dict[str, t.Any] = {}
    assert (
        len(
            list(
                entity_ctx.list_iterator(parameters={"name": file_repository["name"]}, stats=stats)
            )
        )
        == 1
    )
    assert stats["count"] == 1



import pytest

from pulp_glue.common.context import PluginRequirement, PulpContext
from pulp_glue.common.exceptions import PulpException

pytestmark = pytest.mark.glue


def test_has_plugin(mock_pulp_ctx: PulpContext) -> None:
    assert mock_pulp_ctx.has_plugin(PluginRequirement("core"))
    assert mock_pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.50"))
    assert not mock_pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.50,<3.60"))
    assert not mock_pulp_ctx.has_plugin(PluginRequirement("core", specifier=">=3.50,<3.60"))
    assert mock_pulp_ctx.has_plugin(
        PluginRequirement("core", specifier=">=3.50,<3.60", inverted=True)
    )

    assert not mock_pulp_ctx.has_plugin(PluginRequirement("elephant"))
    assert not mock_pulp_ctx.has_plugin(PluginRequirement("elephant", specifier=">=0.0.0"))
    assert mock_pulp_ctx.has_plugin(
        PluginRequirement("elephant", specifier=">=0.0.0", inverted=True)
    )


def test_needs_plugin(mock_pulp_ctx: PulpContext) -> None:
    mock_pulp_ctx.needs_plugin(PluginRequirement("core"))
    mock_pulp_ctx.needs_plugin(PluginRequirement("elephant", feature="zoo"))
    with pytest.raises(PulpException, match=r"elephant.*zoo"):
        mock_pulp_ctx.has_plugin(PluginRequirement("core"))



import random
import string
import typing as t

import pytest

from pulp_glue.common.context import PulpContext
from pulp_glue.file.context import PulpFileRemoteContext, PulpFileRepositoryContext

pytestmark = pytest.mark.glue


@pytest.fixture
def file_remote(pulp_ctx: PulpContext) -> t.Iterator[t.Dict[str, t.Any]]:
    name = "".join(random.choices(string.ascii_letters, k=8))
    file_remote_ctx = PulpFileRemoteContext(pulp_ctx)
    yield file_remote_ctx.create(body={"name": name, "url": "https://localhost/"})
    file_remote_ctx.delete()


@pytest.fixture
def file_repository(pulp_ctx: PulpContext) -> t.Iterator[t.Dict[str, t.Any]]:
    name = "".join(random.choices(string.ascii_letters, k=8))
    file_repository_ctx = PulpFileRepositoryContext(pulp_ctx)
    yield file_repository_ctx.create(body={"name": name})
    file_repository_ctx.delete()


@pytest.mark.live
def test_converge_crud(pulp_ctx: PulpContext) -> None:
    name = "".join(random.choices(string.ascii_letters, k=8))
    description = "".join(random.choices(string.ascii_letters, k=8))
    entity_ctx = PulpFileRepositoryContext(pulp_ctx, entity={"name": name})
    assert entity_ctx.converge({})[0] is True

    entity_ctx.entity = {"name": name}
    assert entity_ctx.converge({})[0] is False

    entity_ctx.entity = {"name": name}
    assert entity_ctx.converge({"description": description})[0] is True

    entity_ctx.entity = {"name": name}
    assert entity_ctx.converge({"description": description})[0] is False

    entity_ctx.entity = {"name": name}
    assert entity_ctx.converge({"description": ""})[0] is True

    entity_ctx.entity = {"name": name}
    assert entity_ctx.converge({"description": ""})[0] is False

    entity_ctx.entity = {"name": name}
    assert entity_ctx.converge(None)[0] is True

    entity_ctx.entity = {"name": name}
    assert entity_ctx.converge(None)[0] is False


@pytest.mark.live
def test_converge_linked_object(
    pulp_ctx: PulpContext, file_repository: t.Dict[str, t.Any], file_remote: t.Dict[str, t.Any]
) -> None:
    name = file_repository["name"]
    entity_ctx = PulpFileRepositoryContext(pulp_ctx, entity={"name": name})
    assert entity_ctx.converge({"remote": file_remote["pulp_href"]})[0] is True

    entity_ctx.entity = {"name": name}
    assert entity_ctx.converge({"remote": file_remote["pulp_href"]})[0] is False

    entity_ctx.entity = {"name": name}
    remote_ctx = PulpFileRemoteContext(pulp_ctx, entity={"name": file_remote["name"]})
    assert entity_ctx.converge({"remote": remote_ctx})[0] is False

    entity_ctx.entity = {"name": name}
    assert entity_ctx.converge({"remote": ""})[0] is True

    entity_ctx.entity = {"name": name}
    assert entity_ctx.converge({"remote": ""})[0] is False


@pytest.mark.live
def test_converge_href(pulp_ctx: PulpContext, file_repository: t.Dict[str, t.Any]) -> None:
    entity_ctx = PulpFileRepositoryContext(pulp_ctx, pulp_href=file_repository["pulp_href"])
    assert entity_ctx.converge({})[0] is False

    entity_ctx.pulp_href = entity_ctx.pulp_href
    assert entity_ctx.converge({"description": "Test converge on repository."})[0] is True

    entity_ctx.pulp_href = entity_ctx.pulp_href
    assert entity_ctx.converge({"description": "Test converge on repository."})[0] is False

    entity_ctx.pulp_href = entity_ctx.pulp_href
    assert entity_ctx.entity["description"] == "Test converge on repository."



import datetime
import io
import typing as t

import pytest

from pulp_glue.common.schema import (
    SchemaError,
    ValidationError,
    encode_json,
    encode_param,
    validate,
)

COMPONENTS = {
    "aString": {"type": "string"},
    "anInteger": {"type": "integer"},
    "aReference": {"$ref": "#/components/schemas/aString"},
    "intArray": {"type": "array", "items": {"type": "integer"}},
    "strArray": {"type": "array", "items": {"$ref": "#/components/schemas/aString"}},
    "minMaxArray": {"type": "array", "minItems": 3, "maxItems": 5, "items": {"type": "integer"}},
    "uniqueArray": {"type": "array", "uniqueItems": True, "items": {"type": "integer"}},
    "unspecifiedObject": {"type": "object"},
    "emptyObject": {"type": "object", "additionalProperties": False},
    "objectAInt": {
        "type": "object",
        "additionalProperties": False,
        "properties": {"a": {"type": "integer"}},
    },
    "objectRequiredA": {
        "type": "object",
        "additionalProperties": {"type": "string"},
        "properties": {
            "a": {"type": "integer"},
            "b": {"type": "boolean"},
        },
        "required": ["a"],
    },
    "allOfEnum": {
        "allOf": [
            {"type": "string", "enum": ["a", "b", "c"]},
            {"type": "string", "enum": ["b", "d", "f"]},
        ]
    },
    "anyOfEnum": {
        "anyOf": [
            {"type": "string", "enum": ["a", "b", "c"]},
            {"type": "string", "enum": ["b", "d", "f"]},
        ]
    },
    "allOfCompose": {
        "allOf": [
            {"type": "object", "properties": {"a": {"type": "string", "format": "date"}}},
            {"type": "object", "properties": {"b": {"type": "string", "format": "date"}}},
        ]
    },
}


# This will only pass for identity...
SOME_BYTES = io.BytesIO(b"\000\001\002")


@pytest.mark.parametrize(
    "schema,value",
    [
        pytest.param({}, None, id="empty_allows_null"),
        pytest.param({}, "asdf", id="empty_allows_string"),
        pytest.param({}, 1, id="empty_allows_int"),
        pytest.param({"nullable": True}, None, id="nullable_allows_null"),
        pytest.param({"type": "string", "nullable": True}, None, id="typed_nullable_allows_null"),
        pytest.param({"type": "string"}, "asdf", id="string"),
        pytest.param({"type": "string", "format": "byte"}, b"\000\001\002", id="string_bytes"),
        pytest.param(
            {"type": "string", "format": "binary"}, b"\000\001\002", id="string_binary_allows_bytes"
        ),
        pytest.param(
            {"type": "string", "format": "binary"}, SOME_BYTES, id="string_binary_allows_bytestream"
        ),
        pytest.param(
            {"type": "string", "format": "date"}, datetime.date(2000, 1, 2), id="string_date"
        ),
        pytest.param(
            {"type": "string", "format": "date"},
            datetime.datetime(2000, 1, 2, 12, 1),
            id="string_date_allows_datetime",
        ),
        pytest.param(
            {"type": "string", "format": "date-time"},
            datetime.datetime(2000, 1, 2, 12, 2),
            id="string_date_allows_datetime",
        ),
        pytest.param({"type": "string", "enum": ["a", "b", "c"]}, "b", id="string_enum"),
        pytest.param({"type": "integer"}, 1, id="integer"),
        pytest.param({"type": "integer", "maximum": 10}, 9, id="integer_max"),
        pytest.param(
            {"type": "integer", "exclusiveMaximum": True, "maximum": 10}, 9, id="integer_exmax"
        ),
        pytest.param({"type": "integer", "maximum": 10}, 10, id="integer_max_1"),
        pytest.param({"type": "integer", "minimum": 5}, 6, id="integer_min"),
        pytest.param(
            {"type": "integer", "exclusiveMinimum": True, "minimum": 5}, 6, id="integer_exmin"
        ),
        pytest.param({"type": "integer", "minimum": 5}, 5, id="integer_min_1"),
        pytest.param({"type": "integer", "maximum": 10, "minimum": 5}, 7, id="integer_minmax"),
        pytest.param({"type": "integer", "multipleOf": 7}, 14, id="integer_multiple_of"),
        pytest.param({"type": "boolean"}, False, id="boolean_false"),
        pytest.param({"type": "boolean"}, True, id="boolean_true"),
        pytest.param({"type": "number"}, 1.5, id="number"),
        pytest.param({"type": "number"}, 1.0, id="whole_number"),
        pytest.param({"type": "number"}, 1, id="number_allows_integer"),
        pytest.param({"$ref": "#/components/schemas/aString"}, "asdf", id="reference"),
        pytest.param({"$ref": "#/components/schemas/aReference"}, "asdf", id="double_reference"),
        pytest.param(
            {"$ref": "#/components/schemas/anInteger", "type": "string"},
            1,
            id="reference_ignores_all_properties",
        ),
        pytest.param({"$ref": "#/components/schemas/intArray"}, [], id="reference_array"),
        pytest.param(
            {"$ref": "#/components/schemas/intArray"}, [1, 1, 2, 3, 5], id="array_of_integer"
        ),
        pytest.param(
            {"$ref": "#/components/schemas/strArray"}, ["a", "b", "c"], id="array_of_strings"
        ),
        pytest.param({"$ref": "#/components/schemas/minMaxArray"}, [1, 1, 3], id="min_max_array"),
        pytest.param(
            {"$ref": "#/components/schemas/unspecifiedObject"}, {}, id="unspecified_object"
        ),
        pytest.param(
            {"$ref": "#/components/schemas/unspecifiedObject"},
            {"a": 1},
            id="unspecified_object_validates_anything",
        ),
        pytest.param({"$ref": "#/components/schemas/emptyObject"}, {}, id="empty_object"),
        pytest.param(
            {"$ref": "#/components/schemas/objectAInt"}, {"a": 1}, id="object_with_properties"
        ),
        pytest.param(
            {"$ref": "#/components/schemas/objectRequiredA"},
            {"a": 1},
            id="object_with_required_property",
        ),
        pytest.param({"$ref": "#/components/schemas/allOfEnum"}, "b", id="all_of"),
        pytest.param(
            {"allOf": [{"type": "date-time"}], "nullable": True},
            None,
            id="all_composes_with_current_schema",
        ),
        pytest.param({"$ref": "#/components/schemas/anyOfEnum"}, "f", id="any_of_matches_one"),
        pytest.param(
            {"$ref": "#/components/schemas/allOfCompose"},
            {"a": datetime.date(2000, 1, 1), "b": datetime.date(2000, 1, 2)},
            id="any_of_matches_one",
        ),
        pytest.param({"type": ["string", "null"]}, "test_string", id="type_array_matches_string"),
        pytest.param({"type": ["string", "null"]}, None, id="type_array_matches_null"),
    ],
)
def test_validates(schema: t.Any, value: t.Any) -> None:
    validate(schema, "testvar", value, COMPONENTS)


@pytest.mark.parametrize(
    "schema,value,match",
    [
        pytest.param({"type": "boolean"}, 1, None, id="boolean_fails_int"),
        pytest.param({"type": "string"}, None, None, id="string_fails_null"),
        pytest.param({"type": "string"}, 1, None, id="string_fails_int"),
        pytest.param({"type": "string"}, True, None, id="string_fails_bool"),
        pytest.param(
            {"type": "string", "nullable": False}, None, None, id="not_nullable_fails_null"
        ),
        pytest.param(
            {"type": "string", "format": "byte"}, "1234", None, id="string_bytes_fails_string"
        ),
        pytest.param(
            {"type": "string", "format": "date"},
            "aritrary string",
            None,
            id="string_date_fails_string",
        ),
        pytest.param(
            {"type": "string", "enum": ["a", "b", "c"]},
            "d",
            None,
            id="string_enum",
        ),
        pytest.param(
            {"type": "integer", "exclusiveMinimum": True, "minimum": 5}, 5, None, id="integer_exmin"
        ),
        pytest.param(
            {"type": "integer", "exclusiveMaximum": True, "maximum": 5}, 5, None, id="integer_exmax"
        ),
        pytest.param({"type": "integer", "minimum": 5}, 4, None, id="integer_min"),
        pytest.param({"type": "integer", "maximum": 5}, 6, None, id="integer_max"),
        pytest.param({"type": "integer", "multipleOf": 7}, 15, None, id="integer_multiple_of"),
        pytest.param(
            {"$ref": "#/components/schemas/aString"}, 1, None, id="string_reference_fails_integer"
        ),
        pytest.param(
            {"type": "number", "minimum": 6.5}, 6.0, None, id="fails_if_number_is_to_small"
        ),
        pytest.param(
            {"$ref": "#/components/schemas/aReference"},
            1,
            None,
            id="double_string_reference_fails_integer",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/intArray",
            },
            True,
            None,
            id="array_fails_boolean",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/intArray",
            },
            ["asdf"],
            None,
            id="integer_array_fails_string_array",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/strArray",
            },
            ["asdf", False, "ghjk"],
            None,
            id="string_array_fails_boolean_in_array",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/minMaxArray",
            },
            [1, 2],
            None,
            id="min_max_array_fails_too_small",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/minMaxArray",
            },
            [1, 2, 3, 4, 5, 6],
            None,
            id="min_max_array_fails_too_large",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/uniqueArray",
            },
            [1, 2, 1, 6],
            None,
            id="unique_array_fails",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/emptyObject",
            },
            [],
            None,
            id="object_fails_list",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/emptyObject",
            },
            {"a": 1},
            None,
            id="object_fails_non_empty",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/objectAInt",
            },
            {"a": "test"},
            None,
            id="object_fails_a_is_wrong_type",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/objectRequiredA",
            },
            {"z": "string"},
            r"'testvar' is missing properties \(a\).",
            id="object_fails_required_property_missing",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/objectRequiredA",
            },
            {"a": 1, "y": 2.5, "z": True},
            r"'testvar\[y\]' is expected to be a string.",
            id="object_fails_additional_property_not_valid",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/allOfEnum",
            },
            "a",
            r"'testvar' is expected to be one of \[b, d, f\].",
            id="fails_validate_only_first_all_of",
        ),
        pytest.param(
            {
                "$ref": "#/components/schemas/anyOfEnum",
            },
            "z",
            None,
            id="fails_validate_none_matched",
        ),
        pytest.param(
            {"type": ["string", "null"]}, 1, "did not match any", id="type_array_matches_string"
        ),
    ],
)
def test_validation_failed(schema: t.Any, value: t.Any, match: t.Optional[str]) -> None:
    with pytest.raises(ValidationError, match=match or r"'testvar.*'"):
        validate(schema, "testvar", value, COMPONENTS)


@pytest.mark.parametrize(
    "schema,value,exc_type",
    [
        pytest.param({"type": "blubb"}, 1, NotImplementedError, id="unknown_type"),
        pytest.param({"$ref": "blubb"}, 1, SchemaError, id="invalid_reference"),
        pytest.param({"$ref": "#/components/schemas/notHere"}, 1, KeyError, id="missing_reference"),
        pytest.param({"type": []}, 1, SchemaError, id="type_array_matches_string"),
    ],
)
def test_invalid_schema_raises(schema: t.Any, value: t.Any, exc_type: t.Type[Exception]) -> None:
    with pytest.raises(exc_type):
        validate(schema, "testvar", value, COMPONENTS)


@pytest.mark.parametrize(
    "obj, output",
    [
        pytest.param(None, "null", id="null"),
        pytest.param(1, "1", id="integer"),
        pytest.param(datetime.date(1970, 1, 1), '"1970-01-01"', id="date"),
        pytest.param(
            datetime.datetime(1970, 1, 1, 12), '"1970-01-01T12:00:00.000000Z"', id="datetime"
        ),
    ],
)
def test_json_encoder(obj: t.Any, output: str) -> None:
    assert encode_json(obj) == output


def test_json_encoder_rejects_binary() -> None:
    with pytest.raises(TypeError):
        encode_json({"a": b"asdf"})


def test_json_encoder_rejects_stream() -> None:
    with pytest.raises(TypeError):
        encode_json({"a": SOME_BYTES})


@pytest.mark.parametrize(
    "value",
    (
        pytest.param("asdf", id="string"),
        pytest.param(42, id="integer"),
    ),
)
def test_encode_param_keeps(value: t.Any) -> None:
    assert encode_param(value) == value


@pytest.mark.parametrize(
    "value,expected",
    (
        pytest.param(datetime.date(2000, 1, 1), "2000-01-01", id="date"),
        pytest.param(
            datetime.datetime(2000, 1, 1, 12, 30), "2000-01-01T12:30:00.000000Z", id="datetime"
        ),
    ),
)
def test_encode_param_transforms(value: t.Any, expected: t.Any) -> None:
    assert encode_param(value) == expected



import typing as t

from pulp_glue.common.context import (
    EntityDefinition,
    PluginRequirement,
    PulpContentContext,
    PulpDistributionContext,
    PulpRemoteContext,
    PulpRepositoryContext,
    PulpRepositoryVersionContext,
)
from pulp_glue.common.i18n import get_translation

translation = get_translation(__package__)
_ = translation.gettext


class PulpAnsibleCollectionVersionContext(PulpContentContext):
    PLUGIN = "ansible"
    RESOURCE_TYPE = "collection_version"
    ENTITY = _("ansible collection version")
    ENTITIES = _("ansible collection versions")
    HREF = "ansible_collection_version_href"
    ID_PREFIX = "content_ansible_collection_versions"
    UPLOAD_ID: t.ClassVar[str] = "upload_collection"
    NEEDS_PLUGINS = [PluginRequirement("ansible", specifier=">=0.7.0")]
    CAPABILITIES = {"upload": [PluginRequirement("ansible", specifier=">=0.16.0")]}

    def upload(self, file: t.IO[bytes], **kwargs: t.Any) -> t.Any:  # type:ignore
        repository: t.Optional[PulpRepositoryContext] = kwargs.pop("repository", None)
        if self.capable("upload"):
            chunk_size: int = kwargs.pop("chunk_size", 1000000)
            return super().upload(file, chunk_size, repository, **kwargs)
        else:
            result = self.call("upload", body={"file": file})
            self.pulp_href = result["created_resources"][0]
            if repository:
                repository.modify(add_content=[self.pulp_href])
            return self.entity


class PulpAnsibleRoleContext(PulpContentContext):
    PLUGIN = "ansible"
    RESOURCE_TYPE = "role"
    ENTITY = _("ansible role")
    ENTITIES = _("ansible roles")
    HREF = "ansible_role_href"
    ID_PREFIX = "content_ansible_roles"
    NEEDS_PLUGINS = [PluginRequirement("ansible", specifier=">=0.7.0")]


class PulpAnsibleCollectionVersionSignatureContext(PulpContentContext):
    PLUGIN = "ansible"
    RESOURCE_TYPE = "collection_version_signature"
    ENTITY = _("ansible collection version signature")
    ENTITIES = _("ansible collection version signatures")
    HREF = _("ansible_collection_version_signature_href")
    ID_PREFIX = "content_ansible_collection_signatures"
    NEEDS_PLUGINS = [PluginRequirement("ansible", specifier=">=0.12.0")]

    def create(
        self,
        body: EntityDefinition,
        parameters: t.Optional[t.Mapping[str, t.Any]] = None,
        non_blocking: bool = False,
    ) -> t.Any:
        self.pulp_ctx.needs_plugin(
            PluginRequirement(
                "ansible", specifier=">=0.13.0", feature=_("collection version creation")
            )
        )
        return super().create(body=body, parameters=parameters, non_blocking=non_blocking)


class PulpAnsibleDistributionContext(PulpDistributionContext):
    PLUGIN = "ansible"
    RESOURCE_TYPE = "ansible"
    ENTITY = _("ansible distribution")
    ENTITIES = _("ansible distributions")
    HREF = "ansible_ansible_distribution_href"
    ID_PREFIX = "distributions_ansible_ansible"
    NEEDS_PLUGINS = [PluginRequirement("ansible", specifier=">=0.7.0")]

    def preprocess_entity(self, body: EntityDefinition, partial: bool = False) -> EntityDefinition:
        body = super().preprocess_entity(body, partial=partial)
        version = body.pop("version", None)
        if version is not None:
            repository_href = body.pop("repository")
            body["repository_version"] = f"{repository_href}versions/{version}/"
        return body


class PulpAnsibleRoleRemoteContext(PulpRemoteContext):
    PLUGIN = "ansible"
    RESOURCE_TYPE = "role"
    ENTITY = _("role remote")
    ENTITIES = _("role remotes")
    HREF = "ansible_role_remote_href"
    ID_PREFIX = "remotes_ansible_role"
    HREF_PATTERN = r"remotes/(?P<plugin>ansible)/(?P<resource_type>role)/"
    NEEDS_PLUGINS = [PluginRequirement("ansible", specifier=">=0.7.0")]


class PulpAnsibleCollectionRemoteContext(PulpRemoteContext):
    PLUGIN = "ansible"
    RESOURCE_TYPE = "collection"
    ENTITY = _("collection remote")
    ENTITIES = _("collection remotes")
    HREF = "ansible_collection_remote_href"
    ID_PREFIX = "remotes_ansible_collection"
    HREF_PATTERN = r"remotes/(?P<plugin>ansible)/(?P<resource_type>collection)/"
    NEEDS_PLUGINS = [PluginRequirement("ansible", specifier=">=0.7.0")]

    def preprocess_entity(self, body: EntityDefinition, partial: bool = False) -> EntityDefinition:
        body = super().preprocess_entity(body, partial=partial)
        if "requirements" in body.keys():
            body["requirements_file"] = body.pop("requirements")
        return body


class PulpAnsibleRepositoryVersionContext(PulpRepositoryVersionContext):
    HREF = "ansible_ansible_repository_version_href"
    ID_PREFIX = "repositories_ansible_ansible_versions"
    NEEDS_PLUGINS = [PluginRequirement("ansible", specifier=">=0.7.0")]


class PulpAnsibleRepositoryContext(PulpRepositoryContext):
    PLUGIN = "ansible"
    RESOURCE_TYPE = "ansible"
    HREF = "ansible_ansible_repository_href"
    ID_PREFIX = "repositories_ansible_ansible"
    ENTITY = _("ansible repository")
    ENTITIES = _("ansible repositories")
    VERSION_CONTEXT = PulpAnsibleRepositoryVersionContext
    CAPABILITIES = {
        "sync": [PluginRequirement("ansible")],
        "pulpexport": [PluginRequirement("ansible")],
    }
    NULLABLES = PulpRepositoryContext.NULLABLES | {"gpgkey"}
    NEEDS_PLUGINS = [PluginRequirement("ansible", specifier=">=0.7.0")]



import typing as t

import pytest
from requests.auth import AuthBase

from pulp_glue.common.exceptions import OpenAPIError
from pulp_glue.common.openapi import AuthProviderBase

pytestmark = pytest.mark.glue


SECURITY_SCHEMES: t.Dict[str, dict[str, t.Any]] = {
    "A": {"type": "http", "scheme": "bearer"},
    "B": {"type": "http", "scheme": "basic"},
    "C": {
        "type": "oauth2",
        "flows": {
            "implicit": {
                "authorizationUrl": "https://example.com/api/oauth/dialog",
                "scopes": {
                    "write:pets": "modify pets in your account",
                    "read:pets": "read your pets",
                },
            },
            "authorizationCode": {
                "authorizationUrl": "https://example.com/api/oauth/dialog",
                "tokenUrl": "https://example.com/api/oauth/token",
                "scopes": {
                    "write:pets": "modify pets in your account",
                    "read:pets": "read your pets",
                },
            },
        },
    },
    "D": {
        "type": "oauth2",
        "flows": {
            "implicit": {
                "authorizationUrl": "https://example.com/api/oauth/dialog",
                "scopes": {
                    "write:pets": "modify pets in your account",
                    "read:pets": "read your pets",
                },
            },
            "clientCredentials": {
                "tokenUrl": "https://example.com/api/oauth/token",
                "scopes": {
                    "write:pets": "modify pets in your account",
                    "read:pets": "read your pets",
                },
            },
        },
    },
}


class MockBasicAuth(AuthBase):
    pass


class MockOAuth2CCAuth(AuthBase):
    pass


def test_auth_provider_select_mechanism(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(AuthProviderBase, "basic_auth", lambda *args: MockBasicAuth())
    monkeypatch.setattr(
        AuthProviderBase,
        "oauth2_client_credentials_auth",
        lambda *args: MockOAuth2CCAuth(),
    )
    provider = AuthProviderBase()

    # Error if no auth scheme is available.
    with pytest.raises(OpenAPIError):
        provider([], SECURITY_SCHEMES)

    # Error if a nonexisting mechanism is proposed.
    with pytest.raises(KeyError):
        provider([{"foo": []}], SECURITY_SCHEMES)

    # Succeed without mechanism for an empty proposal.
    assert provider([{}], SECURITY_SCHEMES) is None

    # Try select a not implemented auth.
    with pytest.raises(OpenAPIError):
        provider([{"A": []}], SECURITY_SCHEMES)

    # Ignore proposals with multiple mechanisms.
    with pytest.raises(OpenAPIError):
        provider([{"B": [], "C": []}], SECURITY_SCHEMES)

    # Select Basic auth alone and from multiple.
    assert isinstance(provider([{"B": []}], SECURITY_SCHEMES), MockBasicAuth)
    assert isinstance(provider([{"A": []}, {"B": []}], SECURITY_SCHEMES), MockBasicAuth)

    # Select oauth2 client credentials alone and over basic auth if scopes match.
    assert isinstance(provider([{"D": []}], SECURITY_SCHEMES), MockOAuth2CCAuth)
    assert isinstance(provider([{"B": []}, {"D": []}], SECURITY_SCHEMES), MockOAuth2CCAuth)
    assert isinstance(
        provider([{"B": []}, {"D": ["read:pets"]}], SECURITY_SCHEMES), MockOAuth2CCAuth
    )
    # Fall back to basic if scope does not match.
    assert isinstance(
        provider([{"B": []}, {"D": ["read:cattle"]}], SECURITY_SCHEMES), MockBasicAuth
    )
